
# $Id: Plugin.pm 231 2008-06-02 12:24:16Z manu $

package PARALOOP::PLUGIN::Plugin;

=pod

=head1 Copyright Notice

 Copyright I.N.R.A. - C.N.R.S. 

 emmanuel.courcelle@toulouse.inra.fr
 jerome.gouzy@toulouse.inra.fr

 This software is a computer program whose purpose is to help you run 
 your scripts on parallel computers.

 This software is governed by the CeCILL  license under French law and
 abiding by the rules of distribution of free software.  You can  use, 
 modify and/ or redistribute the software under the terms of the CeCILL
 license as circulated by CEA, CNRS and INRIA at the following URL
 http://www.cecill.info" 

 As a counterpart to the access to the source code and  rights to copy,
 modify and redistribute granted by the license, users are provided only
 with a limited warranty  and the software's author,  the holder of the
 economic rights,  and the successive licensors  have only  limited
 liability. 

 In this respect, the user's attention is drawn to the risks associated
 with loading,  using,  modifying and/or developing or reproducing the
 software by the user in light of its specific status of free software,
 that may mean  that it is complicated to manipulate,  and  that  also
 therefore means  that it is reserved for developers  and  experienced
 professionals having in-depth computer knowledge. Users are therefore
 encouraged to load and test the software's suitability as regards their
 requirements in conditions enabling the security of their systems and/or 
 data to be ensured and,  more generally, to use and operate it in the 
 same conditions as regards security. 

 The fact that you are presently reading this means that you have had
 knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

PARALOOP::PLUGIN::Plugin - Abstract class for the Plugin objects

=head1 Derives from...

UNIVERSAL 

=head1 DESCRIPTION

The objects deriving from Plugin implement 5 methods, requested by paraloop. Plugin calls dies
for some of those, thus declaring abstract methods.

=head1 PUBLIC METHODS


=head2 I/O METHODS (all abstract)

 NextRecord
    $obj->NextRecord();

 SkipRecords
    $obj->SkipRecords(10)

 Tell
    my $record_number = $obj->Tell();

 Seek
    my $obj->Seek(10);

 TellLength
    my $nb_of_records = $obj->TellLength();

=head2 OTHER METHODS

 WhatParaloopPlugin
    my $msg=$obj->WhatParaloopPlugin();
    ABSTRACT

 Parameters
    my $msg=$obj->Parameters();
    ABSTRACT
    
 New
    my $obj = PARALOOP::Dummy->New(-params => $params, -log => $log);
    ABSTRACT (cf _Init)

 Exec
    my ($incr_size,$error,$killed) = $obj->Exec();
    ABSTRACT

 SetOutputFile
    $obj->SetOutputFile('filename');
    
 __JobId
    my $job_id = $obj->__JobId($o_param);

 StealWork (for load balancing)
    $obj->StealWork($o_scheduler);
    
=head1 METHODS AND SUBROUTINES - DETAILED DESCRIPTION

=cut

use strict;
use warnings;
use Cwd;
use File::Basename;
use Carp;

require "UNIVERSAL.pl";
use ParamParser;
use Logger;
use Runner;
use Fcntl;

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 231 $ =~ /\d+/g); $r[0]};

}

=pod

=head2 Function WhatParaloopPlugin

 Usage:    my $msg = PARALOOP::Dummy::WhatParaloopPlugin()
        or my $msg = $obj->WhatParaloopPlugin()
 Function: Return some information (2 lines max) on the plugin
           An object is NOT considered as a plugin if this function is not implemented
 Arg:      none ($self is ignored)
 Return:   A message describing the plugin (1 or 2 lines)

=cut

sub WhatParaloopPlugin {die "\n";}

=pod

=head2 Function Parameters

 Usage:    my $msg = PARALOOP::Dummy::Parameters()
        or my $msg = $obj->Parameters()
 Function: Return the list of allowed parameters for this plugin
           If your plugin defines some parameters, it is good pratice adding this function,
           to display those parameters when the --parameters switch of paraloop.pl is used.
           However, this is not mandatory
 Arg:      none ($self is ignored)
 Return:   A message describing the plugin (1 or 2 lines)

=cut

sub Parameters {die "\n";}

=pod

=head2 procedure _Init

 Usage:  CALLED BY UNIVERSAL::New
         my obj = PARALOOP::Dummy->New(-params => $params, -log => $log);
 Access: private (use New instead)
 procedure: Init the object
 Return:   nothing
 Args:     -params = an object of class ParamParser (already initialized)
           -log    = an object of class Log (already initialized)
 globals:  none

=cut

sub _Init {
    my $self = shift;
    #print `hostname`." Plugin.pm ".join(' ',@_)."\n";
    my %args = @_;
    my $o_param = $args{'-params'};
    my $o_log   = $args{'-log'};
    
    my $tmp;
    if ($o_param->IsDefined('tmp'))
    { 
        $tmp = $o_param->Get('tmp');
    }
    else
    {
        $tmp = '/tmp';
    };

    my $stamp = $o_log->CalcStamp();
    my $job_id= $self->_JobId($o_param);
    my $test_file = $tmp . '/' . 'test.out.'.$job_id.'.'.$stamp;
    open(TEST, ">$test_file") or croak("$test_file cannot be created, you should check the PARALOOP_tmp parameter");
    close(TEST);
    unlink($test_file);

    $self->{'__tmp'} = $tmp;
    
    my $log_level = $o_param->Get('log_level');
    my $rewrite_input = ($log_level =~ /R/);
    $self->{'__rewrite_input'} = $rewrite_input;
}

=pod

=head2 Procedure SetOutputFile

 Usage:     &$opt->SetOutputFile('filename');
 Procedure: change the name of the output file
 Arg :      the output file name

=cut

sub SetOutputFile
{
    my $self       = shift;
    my $outputfile = shift;
    $self->{'__outputfile'} = $outputfile;
}

=head2 Function Exec

 Usage:    my ($incr_size, $error,$killed) = &$opt->Exec($timeout);
 Function: Abstract method - What is done depends on the derived object
 Arg:      $timeout   = The number of seconds after which a timeout is sent
 Return:   $incr_size = The number of bytes added to the output file
           $error     = An error code. 
		                If != 0, paraloop WILL BE TERMINATED - Generally marks the end of file
           $killed    = A kill code
		                0: everything OK 
                       -1: killed by timeout
                        -2: killed by TERM signal (kill, qdel, ...)

=cut

sub Exec {die "\n";}

=head2 procedure CallExternalScript

 Usage: $obj->CallExternalScript($prm,$input,$output)
 procedure: Call an external script whose path is stored in $o_param (parameter $prm)
            If the parameter does not exist and if $output not defined, does nothing
            If the parameter does not exist BUT $output starts with '>', truncate $output to 0: the caller is sure that
            the $output file will be removed even if there is no external script to call (without this trick the Shell plugin, among others,
            did not truncate the temporary output file, and the final output file did reach huge sizes)
              
 Return:    Nothing
 Args:      $prm The parameter to get (one of preproc_script, postproc_script, preatom_script, postatom_script)
            $input (optional)  The input file
            $output (optional) The output file (must start with > or >>)

=cut

sub CallExternalScript
{
    my $self = shift;
    my $prm = shift;
    my $input = shift;
    my $output= shift;

    my $o_param = $self->{'__params'};
    my $o_log   = $self->{'__log'};

    if ($o_param->IsDefined($prm))
    {
        my $cmd = $o_param->Get($prm);
        my $path = $cmd;
        $path =~ s/(.*) +.*/$1/; 
        if ( ! -x $path)
        {
            $o_log->Trace("CANNOT RUN EXTERNAL SCRIPT $path NOT FOUND OR NOT EXECUTABLE\n",'.','0');
        }
        else
        {
            my %args;
            $args{-cmd}=$cmd;
            $args{-stdin} = $input if (defined $input);
            $args{-stdout}= $output if (defined $output);
            $o_log->StartTimer($prm);
            my ($sts, $kld) = RunExt(%args);
            $o_log->Trace("CALLING EXTERNAL SCRIPT $prm - Return $sts\n",$prm,'1');
        }
    }
    else
    {
        if (defined($output) and $output =~ /^ *>[^>]/)
        {
            my $trucated = new IO::File($output);
        }
    }
}
    
=pod

=head2 function NextRecord

 Usage:    my $line = $obj->NextRecord()
 function: read and return next record (a string)
           update private member __records_counter
 Return:   a string
 Args:     none

=cut

sub NextRecord {die "\n";}

=pod

=head2 function Tell

 Usage:    my $cnt = $obj->Tell()
 function: return records_counter private field
 Return:   The counter of records (first=0)
 Args:     none

=cut

sub Tell {die "\n";}

=pod

=head2 function TellLength

 Usage:    my $cnt = $obj->TellLength()
 function: Open the file again, call next_seq() until the end the file is reached, 
           counting the number of records in the file.
 Return:   The total number of records
 Args:     none

=cut

sub TellLength {die "\n";}

=pod

=head2 procedure Seek

 Usage:    $obj->Seek($n)
 procedure:Position the pointer to $n (absolute, 1st record is 0)
 Args:     The record to seek to (MUST BE > 0)

=cut

sub Seek {die "\n";}

=pod

=head2 procedure SkipRecords

 Usage:    $obj->SkipRecords($n)
 procedure:Position the pointer, skipping $n records
 Args:     The number of records to skip (MUST BE > 0)
 Access:   protected

=cut

sub SkipRecords {die "\n";}

=head2 function _JobId

 Usage:    my $job_id = $obj->__JobId($o_param)
 procedure:Return the job_id, as found in the ParamParser object.
           If 0, return a pseudo job_id, just good enough to create a temporary directory of file
 Args:     A ref to a ParamParsers object
 Access:   protected

=cut

sub _JobId
{
    my $self    = shift;
    my $o_param = shift;
    my $job_id  = $o_param->IsDefined('job_id') ? $o_param->Get('job_id') : int(rand 1000000000);
    return $job_id;
}

=pod

=head2 Function StealWork

 Usage: my $sts = $obj->StealWork($o_scheduler,$threshold)
 function:  Look inside all the jobs, trying to steal some work to another job.
            1/ Look for the jobs whose status is anything but TERMINATED or INTERRUPTED
            2/ Select the less advanced job
            3/ If this job has a more than 1 record to work with (configurable with PARALOOP_load_balancing_threshold parameter), submit a new
               monoprocessor job, taking 50% of the remaining work
               This function may be overloaded by the plugins, if necessary.
 precondition: $o_scheduler must exist and be initialized
 Return:   0 if nothing stolen, 1 if something stolen
 Args:     $o_scheduler The scheduler
           $threshold   Optional argument. If present, use this threshold (may be useful for subclasses)
                                           If not present, look for threshold in $o_param
 Globals:  none

=cut

sub StealWork
{
    my $self = shift;
    my ($o_sched,$threshold) = (@_);

    my $o_params = $self->{'__params'};
    my $o_log    = $self->{'__log'};

    if (!defined $threshold)
    {    
        $threshold = ($o_params->IsDefined('load_balancing_threshold') ? $o_params->Get('load_balancing_threshold') : main::DEF_LOAD_BALANCING_THRESHOLD);
    }
    
    my $o_stolen_params;
    my $stolen_remaining_work=0;
    
    my $lock_file= $o_params->Get('lock_file');
    my $lock_dir = dirname($lock_file);

    # Lock the whole directory, because we must be the only stealer    
    my $semaphore= $lock_dir.'/.steallock';
    open(L,"+>>$semaphore") or die "Cannot open $semaphore";
    fcntl(L,F_SETLKW,pack('ssx32',F_WRLCK,0)) or die "Can't put a write lock on $semaphore: $!";

    foreach my $f (glob("$lock_dir/lock.*.lck"))
    {
        # Avoid stealing something to myself
        next if ($lock_file eq $f);         
        
        my $o_f_params = New ParamParser($f);
        $o_f_params->Update($f.'_tmp','O') if (-e $f.'_tmp'); 
        $o_f_params->SelectNameSpace(main::NAMESPACE);
        
        #The master has nothing to do: don't steal anything to him
        next if ($o_f_params->IsDefined('current_mode') and $o_f_params->Get('current_mode') eq 'MASTER');
        
        # Ignore job if terminated or interrupted
        next if ($o_f_params->Get('status') eq 'TERMINATED' or $o_f_params->Get('status') eq 'INTERRUPTED');

        my $f_start   = $o_f_params->Get('start') + 0;
        my $f_end     = $o_f_params->Get('end') + 0;
        
        # NOTE 'current' points to an atomic job that is ALREADY computed and saved
        #      thus, we must increment it to point to the REAL current atomic job
        my $f_current = $o_f_params->Get('current') + 1;
        
        my $f_remaining_work = $f_end - $f_current;
        next if ($f_remaining_work <= $threshold);           # Few steps - Don't steal this guy, he does not have enough work

        if ($stolen_remaining_work < $f_remaining_work)
        {
            $stolen_remaining_work = $f_remaining_work;
            $o_stolen_params = $o_f_params;
        };
    };
    
    #
    # We found some work to steal: create the corresponding tmp lock -file, and submit the new job
    #
    
    my $something_stolen=0;
    if ($stolen_remaining_work > 0)
    {
        #print "FOUND\n";
        my $stolen_current = $o_stolen_params->Get('current');
        
        my $stolen_end     = $o_stolen_params->Get('end');
        my $stealer_end    = $stolen_end;
        
        my $stolen_nb_of_steps = int((1 + $stolen_end - $stolen_current) / 2); # Steal only half of the remaining work
        my $stealer_nb_of_steps= 1 + $stolen_end - $stolen_current - $stolen_nb_of_steps;
        
        if ($stolen_nb_of_steps > 0 and $stealer_nb_of_steps > 0 )      # Nothing to steal is one of them == 0
        {
            my $stolen_lock_file   = $o_stolen_params->Get('lock_file');
            my $stolen_tmp_lock_file = $stolen_lock_file . '_tmp';
            #print "KOUKOU $stolen_tmp_lock_file ".$o_stolen_params->Get('current_mode')."\n";
            my $stolen_cmd           = $o_stolen_params->Get('command_line');
            my $stolen_job_id        = $o_stolen_params->Get('job_id');
            
            # Modify $stolen_end, because some work is stolen
            $stolen_end = $stolen_current + $stolen_nb_of_steps - 1;
            
            # Init $o_stolen_params and recreate a params object with only the changed parameters
            # Save this object to $stolen_tmp_lock_file
            
            $o_stolen_params->Init();
            $o_stolen_params->SelectNameSpace(main::NAMESPACE);
            $o_stolen_params->Set('end',$stolen_end);
            $o_stolen_params->Dump($stolen_tmp_lock_file,main::NAMESPACE);
            
            # Compute $stealer_start
            my $stealer_start = $stolen_current +  $stolen_nb_of_steps;
            
            # Store $stealer_start, $stealer_end in $o_params, keeping old values
            
            my $start = $o_params->Get('start');
            my $end   = $o_params->Get('end');
            my $step  = 1;
            
            my $done_segments      = $o_params->Get('done_segments');
            my $done_time_elapsed  = $o_params->Get('done_time_elapsed');
            
            if ($done_segments eq '')
            {
                $done_segments     = $start.'-'.$end.'-'.$step;
                $done_time_elapsed = $o_params->Get('time_elapsed');
            }
            else
            {
                $done_segments     .= ','.$start.'-'.$end.'-'.$step;
                $done_time_elapsed .= ','.$o_params->Get('time_elapsed');
            };
            
            $o_params->Set('done_segments',$done_segments);
            $o_params->Set('done_time_elapsed',$done_time_elapsed);
            $o_params->Set('time_elapsed',0);
            $o_params->Set('start',$stealer_start);
            $o_params->Set('current',$stealer_start);
            $o_params->Set('end',$stealer_end);
            
            $self->Seek(0);     # Rewind the input file !!!
            
            $o_log->Trace("LOAD BALANCING: NOW STEALING $stealer_start .. $stealer_end (threshold $threshold) TO $stolen_job_id (see $stolen_lock_file) - \n",'*','1');
            $o_log->Trace("$stolen_job_id RECONFIGURED: NEW END $stolen_end (CURRENT $stolen_current)",'*','2');
            
            $something_stolen = 1;
            
        }
    }

    # Unlock the directory    
    fcntl(L,F_SETLKW,pack('ssx32',F_UNLCK,0)) or die "Can't release the write lock on $semaphore: $!";
    close(L);
    
    return $something_stolen;
}

=pod

=head2 procedure DESTROY

 Usage:     The destructor of the object
            undef($obj);
 Access:    public
 procedure: Destroyes the resources allocated by _Init
 Return:    nothing
 Args:      nothing
 globals:   none

=cut

sub DESTROY
{
    my $self   = shift;
    return unless (exists $self->{'__params'});
    my $o_param=$self->{'__params'};
    my $rm = $o_param->Get('rm');
    
    my $tmpdir = $self->{'__tmpdir'};
    `$rm -rf $tmpdir` if (defined $tmpdir);
}

1;

